import logging

# Base class for a Thing which has a name and traits.
from soakdb3_api.thing import Thing

logger = logging.getLogger(__name__)


class Base(Thing):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, thing_type, specification=None, predefined_uuid=None):
        Thing.__init__(self, thing_type, specification, predefined_uuid=predefined_uuid)

        self.__logging_mpqueue = None

    # ----------------------------------------------------------------------------------------
    def set_logging_mpqueue(self, mpqueue):
        self.__logging_mpqueue = mpqueue

    def get_logging_mpqueue(self):
        return self.__logging_mpqueue



# ********** Please don't edit this file!
# ********** It has been generated automatically by dae_devops version 0.5.1.
# ********** For repository_name soakdb3

from pathlib import Path
from subprocess import check_output

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from sphinx.domains.python import PythonDomain

import soakdb3_lib

# -- General configuration ------------------------------------------------

# General information about the project.
project = "soakdb3"

# The full version, including alpha/beta/rc tags.
release = soakdb3_lib.__version__

# The short X.Y version.
if "+" in release:
    # Not on a tag, use branch name
    root = Path(__file__).absolute().parent.parent
    git_branch = check_output("git branch --show-current".split(), cwd=root)
    version = git_branch.decode().strip()
else:
    version = release

extensions = [
    # Use this for generating API docs
    "sphinx.ext.autodoc",
    # This can parse google style docstrings
    "sphinx.ext.napoleon",
    # For linking to external sphinx documentation
    "sphinx.ext.intersphinx",
    # Add links to source code in API docs
    "sphinx.ext.viewcode",
    # Adds the inheritance-diagram generation directive
    "sphinx.ext.inheritance_diagram",
    # Add a copy button to each code block
    "sphinx_copybutton",
    # For the card element
    "sphinx_design",
    # For command line tools autodoc.
    "sphinxarg.ext",
    # Create pages from jupyter notebooks
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
]

# If true, Sphinx will warn about all references where the target cannot
# be found.
nitpicky = True

# A list of (type, target) tuples (by default empty) that should be ignored when
# generating warnings in "nitpicky mode". Note that type should include the
# domain name if present. Example entries would be ('py:func', 'int') or
# ('envvar', 'LD_LIBRARY_PATH').
nitpick_ignore = [("py:class", "numpy.ma.core.MaskedArray")]

# Workaround for NewType as autodata, to be removed when issue is resolved
# see: https://github.com/sphinx-doc/sphinx/issues/9560
assert PythonDomain.object_types["data"].roles == ("data", "obj")
PythonDomain.object_types["data"].roles = ("data", "class", "obj")
# Both the class’ and the __init__ method’s docstring are concatenated and
# inserted into the main body of the autoclass directive
autoclass_content = "both"

# Order the members by the order they appear in the source code
autodoc_member_order = "bysource"

# Don't inherit docstrings from baseclasses
autodoc_inherit_docstrings = False

# Don't show the typehints in the function/method signature.
autodoc_typehints = "description"

# Output graphviz directive produced images in a scalable format
graphviz_output_format = "svg"

# The name of a reST role (builtin or Sphinx extension) to use as the default
# role, that is, for text marked up `like this`
default_role = "any"

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# These patterns also affect html_static_path and html_extra_path
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# This means you can link things like `str` and `asyncio` to the relevant
# docs in the python documentation.
intersphinx_mapping = dict(
    python=("https://docs.python.org/3/", None),
    numpy=("https://numpy.org/doc/stable/", None),
)

# A dictionary of graphviz graph attributes for inheritance diagrams.
inheritance_graph_attrs = dict(rankdir="TB")

# Common links that should be available on every page
rst_epilog = """
.. _Diamond Light Source: http://www.diamond.ac.uk
.. _black: https://github.com/psf/black
.. _flake8: https://flake8.pycqa.org/en/latest/
.. _isort: https://github.com/PyCQA/isort
.. _mypy: http://mypy-lang.org/
.. _pre-commit: https://pre-commit.com/
"""

# Ignore localhost links for periodic check that links in docs are valid
linkcheck_ignore = [r"http://localhost:\d+/"]

# Set copy-button to ignore python and bash prompts
# https://sphinx-copybutton.readthedocs.io/en/latest/use.html#using-regexp-prompt-identifiers
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# These folders are copied to the documentation's HTML output
html_static_path = ["_static"]

# Paths are either relative to html_static_path or fully qualified paths (eg. https://...)
html_css_files = [
    # Custom css to allow use of full window width in the browser.
    "css/custom.css",
]

# Theme options for pydata_sphinx_theme
html_theme_options = dict(
    logo=dict(
        text=project,
    ),
    gitlab_url="None/soakdb3",
    icon_links=[],
    navbar_end=["theme-switcher", "icon-links"],
)

# A dictionary of values to pass into the template engine’s context for all pages
html_context = dict(
    doc_path="docs",
)

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# Disable the link to show the rst source.
# I did this since it's noise on the page and most audience doesn't care to see the raw rst.
html_show_sourcelink = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Logo
html_logo = "images/dls-logo.svg"
html_favicon = "images/dls-favicon.ico"


def ultimateReplace(app, docname, source):
    result = source[0]
    for key in app.config.ultimate_replacements:
        result = result.replace(key, app.config.ultimate_replacements[key])
    source[0] = result


# I got this from https://github.com/sphinx-doc/sphinx/issues/4054.
# It will allow the ${token} replacement in the rst documents.
ultimate_replacements = {
    "$" + "{repository_name}": "soakdb3",
    "$" + "{package_name}": "soakdb3_lib",
    "$" + "{git_url}": "None",
    "$" + "{python_version_at_least}": "3.1",
}


def setup(app):
    app.add_config_value("ultimate_replacements", {}, True)
    app.connect("source-read", ultimateReplace)


# dae_devops_fingerprint 240d703cf524d0fc6ca78c9139863acf



import asyncio

# Use standard logging in this module.
import logging

# Utilities.
from dls_utilpack.describe import describe

# Base class for cli subcommands.
from soakdb3_cli.subcommands.base import ArgKeywords, Base

# Context creator.
from soakdb3_lib.contexts.contexts import Contexts

logger = logging.getLogger()

# Specifications of services we can start, and their short names for parse args.
services = {
    "dataface_specification": "dataface",
}


# --------------------------------------------------------------
class StartServices(Base):
    """
    Start one or more services and keep them running until ^C.
    """

    def __init__(self, args, mainiac):
        super().__init__(args)

    # ----------------------------------------------------------------------------------------
    def run(self):
        """ """

        # Run in asyncio event loop.
        asyncio.run(self.__run_coro())

    # ----------------------------------------------------------
    async def __run_coro(self):
        """"""

        # Load the configuration.
        configurator = self.get_configurator(vars(self._args))

        # Let the configurator know about any mpqueue logging.
        # configurator.set_logging_mpqueue(self.__mainiac.mpqueue)

        context_configuration = await configurator.load()

        if len(self._args.service_names) == 0:
            self._args.service_names = ["all"]

        if "all" in self._args.service_names:
            selected_service_names = []
            for _, service_name in services.items():
                selected_service_names.append(service_name)
        else:
            selected_service_names = self._args.service_names

        logger.debug(describe("selected_service_names", selected_service_names))

        # Change all start_as to None, except the one we are starting.
        for keyword, specification in context_configuration.items():
            if keyword in services:
                service_name = services[keyword]
                if service_name in selected_service_names:
                    specification["context"] = {"start_as": "process"}

        # Make a context from the configuration.
        context = Contexts().build_object(context_configuration)

        # Open the context (servers and clients).
        async with context:

            # Announce.
            logger.info("services started, waiting for excel clients to connect...")

            try:
                # Stay up until all processes are dead.
                # TODO: Use asyncio wait or sentinel for all started processes to be dead.
                while True:
                    await asyncio.sleep(1.0)
                    if not await context.is_any_process_alive():
                        logger.info("all processes have shutdown")
                        break
            except KeyboardInterrupt:
                pass

    # ----------------------------------------------------------
    def add_arguments(parser):

        services_list = list(services.values())

        parser.add_argument(
            help='"all" (default) or any combination of {%s}'
            % (" ".join(services_list)),
            nargs="*",
            type=str,
            metavar="service name(s)",
            dest="service_names",
            default=[],
        )

        parser.add_argument(
            "--configuration",
            "-c",
            help="Configuration file.",
            type=str,
            metavar="yaml filename",
            default=None,
            dest=ArgKeywords.CONFIGURATION,
        )

        return parser



import logging

# Utilities.
from dls_utilpack.callsign import callsign

# Base class for a Thing which has a name and traits.
from soakdb3_api.thing import Thing

logger = logging.getLogger(__name__)


class Base(Thing):
    """ """

    # ----------------------------------------------------------------------------------------
    def __init__(self, thing_type, specification=None, predefined_uuid=None):
        Thing.__init__(self, thing_type, specification, predefined_uuid=predefined_uuid)

        # Reference to object which is a server, such as BaseAiohttp.
        self.server = None

        self.context_specification = self.specification().get("context", {})

    # ----------------------------------------------------------------------------------------
    async def is_process_started(self):
        """"""

        if self.server is None:
            raise RuntimeError(f"{callsign(self)} a process has not been defined")

        try:
            return await self.server.is_process_started()
        except Exception:
            raise RuntimeError(
                f"unable to determing process started for server {callsign(self.server)}"
            )

    # ----------------------------------------------------------------------------------------
    async def is_process_alive(self):
        """"""

        if self.server is None:
            raise RuntimeError(f"{callsign(self)} a process has not been defined")

        try:
            return await self.server.is_process_alive()
        except Exception:
            raise RuntimeError(
                f"unable to determing dead or alive for server {callsign(self.server)}"
            )

    # ----------------------------------------------------------------------------------------
    async def __aenter__(self):
        """ """

        await self.aenter()

    # ----------------------------------------------------------------------------------------
    async def __aexit__(self, type, value, traceback):
        """ """

        await self.aexit()



#!/usr/bin/env python

import argparse

# Use standard python logging
import logging
import multiprocessing

# Base class with methods supporting MaxIV command-line programs.
from dls_mainiac_lib.mainiac import Mainiac

# The subcommands.
from soakdb3_cli.subcommands.start_services import StartServices

# The package version.
from soakdb3_cli.version import meta as version_meta
from soakdb3_cli.version import version

logger = logging.getLogger(__name__)


# --------------------------------------------------------------
class Main(Mainiac):
    def __init__(self, app_name):
        super().__init__(app_name)

    # ----------------------------------------------------------
    def run(self):
        """"""

        if self._args.subcommand == "start_services":
            StartServices(self._args, self).run()

        else:
            raise RuntimeError("unhandled subcommand %s" % (self._args.subcommand))

    # ----------------------------------------------------------
    def build_parser(self, arglist=None):
        """
        Method called from mainiac command line parsing.
        Should return argparser for this program.
        """

        # Make a parser.
        parser = argparse.ArgumentParser(
            description="Command line app for checking quality of femtoscan file in progress.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )

        # --------------------------------------------------------------------

        parser.add_argument(
            "--log_name",
            help="basis name for log",
            type=str,
            metavar="string",
            dest="log_name",
            default=None,
        )

        # --------------------------------------------------------------------
        subparsers = parser.add_subparsers(help="subcommands", dest="subcommand")
        subparsers.required = True

        # --------------------------------------------------------------------
        subparser = subparsers.add_parser("start_services", help="Start service(s).")
        StartServices.add_arguments(subparser)

        return parser

    # --------------------------------------------------------------------------
    def configure_logging(self, settings=None):
        """
        Configure runtime logging, override base class.
        Presume that self._args is already set.
        """

        if self._args.log_name is None:
            self._args.log_name = self._args.subcommand

        # Name as it appears in logging.
        multiprocessing.current_process().name = self._args.log_name

        # Set mainaic's program name to include the subcommand.
        self.program_name("%s/%s" % (self.program_name(), self._args.log_name))

        # Enable the multiprocessing queue listener.
        settings = {
            "mpqueue": {"enabled": True},
        }

        # Let the base class do most of the work.
        Mainiac.configure_logging(self, settings)

        # See bisstis-maxiv-daqcluster for advanced logging filters.

        # Don't show matplotlib font debug.
        logging.getLogger("matplotlib.font_manager").setLevel("INFO")

        # Set filter on the ispyb logger to ignore the annoying NOTICE.
        logging.getLogger("ispyb").addFilter(_ispyb_logging_filter())

    # ----------------------------------------------------------
    def version(self):
        """
        Method called from mainiac command line parsing.
        Should return string in form of N.N.N.
        """
        return version()

    # ----------------------------------------------------------
    def about(self):
        """
        Method called from mainiac command line parsing.
        Should return dict which can be serialized by json.
        """

        return {"versions": version_meta()}


# --------------------------------------------------------------------------------
class _ispyb_logging_filter:
    """
    Python logging filter to remove annoying traitlets messages.
    These are not super useful to see all the time at the DEBUG level.
    """

    def filter(self, record):

        if record.msg.startswith(
            "NOTICE: This code uses __future__ functionality in the ISPyB API."
        ):
            return 0

        return 1


# # --------------------------------------------------------------------------------
# class _matplotlib_logging_filter:
#     """
#     Python logging filter to remove annoying matplotlib messages.
#     These are not super useful to see all the time at the INIT level.
#     """

#     def filter(self, record):
#         if "loaded modules" in record.msg:
#             return 0

#         return 1


# ---------------------------------------------------------------
def main():

    # Instantiate the app.
    main = Main("soakdb3_cli")

    # Configure the app from command line arguments.
    main.parse_args_and_configure_logging()

    # Run the main wrapped in a try/catch.
    main.try_run_catch()


# ---------------------------------------------------------------
# From command line, invoke the main method.
if __name__ == "__main__":
    main()



import logging

# Base class for the aiosqlite database object.
from dls_normsql.aiosqlite import Aiosqlite as NormsqlAiosqlite

from soakdb3_api.databases.constants import Tablenames

# Base class for our database definition.
from soakdb3_lib.databases.database_definition import DatabaseDefinition

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class Aiosqlite(DatabaseDefinition, NormsqlAiosqlite):
    """
    Class with coroutines for creating and querying a sqlite database.
    We use dls-normsql to do the heavy lifting.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        """
        Construct object.  Do not connect to database.
        """

        # Constructor for the database definition.
        DatabaseDefinition.__init__(self)

        # Constructor for the database implementation.
        NormsqlAiosqlite.__init__(self, specification)

    # ----------------------------------------------------------------------------------------
    def reinstance(self):
        """
        Clone database instance.
        This is needed if a process inherits an instance, but needs its own connection.
        """

        return Aiosqlite(self.__filename)

    # ----------------------------------------------------------------------------------------
    async def add_table_definitions(self):
        """
        Make all the table definitions.
        """

        # Add tables common in all implementations.
        await NormsqlAiosqlite.add_table_definitions(self)

        # Add tables from our definition.
        await DatabaseDefinition.add_table_definitions(self)

    # ----------------------------------------------------------------------------------------
    async def apply_revision(self, revision):

        # Let the base class add any common updates.
        # Usually only does anything if upgrading to revision 1
        # which means we are starting with a legacy database with no revision information.
        await NormsqlAiosqlite.apply_revision(self, revision)

        # Updating to revision 1 presumably means
        # this is a legacy database with no revision table in it.
        if revision == 1:
            await self.create_table(Tablenames.VISIT)

        # if revision == 2:
        #     await self.execute(
        #         f"ALTER TABLE {Tablenames.ROCKMAKER_IMAGES} ADD COLUMN {ImageFieldnames.NEWFIELD} TEXT",
        #         why=f"revision 2: add {Tablenames.ROCKMAKER_IMAGES} {ImageFieldnames.NEWFIELD} column",
        #     )
        #     await self.execute(
        #         "CREATE INDEX %s_%s ON %s(%s)"
        #         % (
        #             Tablenames.ROCKMAKER_IMAGES,
        #             ImageFieldnames.NEWFIELD,
        #             Tablenames.ROCKMAKER_IMAGES,
        #             ImageFieldnames.NEWFIELD,
        #         )
        #     )



import logging
import os
import shutil

import pytest

# Formatting of testing log messages.
from dls_logformatter.dls_logformatter import DlsLogformatter

logger = logging.getLogger(__name__)


# --------------------------------------------------------------------------------
@pytest.fixture(scope="session")
def constants(request):

    constants = {}

    yield constants


# --------------------------------------------------------------------------------
@pytest.fixture(scope="session")
def logging_setup():

    formatter = DlsLogformatter(type="long")
    logger = logging.StreamHandler()
    logger.setFormatter(formatter)
    logging.getLogger().addHandler(logger)

    # Log level for all modules.
    logging.getLogger().setLevel("DEBUG")

    # Turn off noisy debug.
    logging.getLogger("asyncio").setLevel("WARNING")
    logging.getLogger("pika").setLevel("WARNING")
    logging.getLogger("stomp").setLevel("WARNING")
    logging.getLogger("luigi-interface").setLevel("WARNING")
    logging.getLogger("luigi.xchem_be_scheduler").setLevel("INFO")
    logging.getLogger("urllib3.connectionpool").setLevel("INFO")

    logging.getLogger("soakdb3_lib.things").setLevel("INFO")

    # Messages about starting and stopping services.
    logging.getLogger("soakdb3_lib.base_aiohttp").setLevel("INFO")

    # All xchem_be database sql commands.
    # logging.getLogger("soakdb3_lib.databases.aiosqlite").setLevel("INFO")

    logging.getLogger("soakdb3_lib.contexts.classic").setLevel("INFO")
    logging.getLogger("soakdb3_lib.datafaces.context").setLevel("INFO")

    # Registering signal handler.
    logging.getLogger("dls_siggy_lib.signal").setLevel("INFO")

    # Set filter on the ispyb logger to ignore the annoying NOTICE.
    logging.getLogger("ispyb").addFilter(_ispyb_logging_filter())

    # Cover the version.
    # logger.info("\n%s", (json.dumps(version_meta(), indent=4)))

    yield None


# --------------------------------------------------------------------------------
class _traitlets_logging_filter:
    """
    Python logging filter to remove annoying traitlets messages.
    These are not super useful to see all the time at the DEBUG level.
    """

    def filter(self, record):

        if record.levelno == 10:
            if "jupyter_client/client.py" in record.pathname:
                return 0
            if "jupyter_client/connect.py" in record.pathname:
                return 0
            if "jupyter_client/manager.py" in record.pathname:
                return 0
            if "jupyter_client/provisioning/factory.py" in record.pathname:
                return 0
            if "nbclient/client.py" in record.pathname:
                return 0
            if "nbconvert/exporters/templateexporter.py" in record.pathname:
                return 0
            if "nbconvert/preprocessors/base.py" in record.pathname:
                return 0
            if "/nbconvert/preprocessors/coalescestreams.py" in record.pathname:
                return 0

            # if "" in record.pathname:
            #     return 0

        return 1


# --------------------------------------------------------------------------------
class _ispyb_logging_filter:
    """
    Python logging filter to remove annoying traitlets messages.
    These are not super useful to see all the time at the DEBUG level.
    """

    def filter(self, record):

        if record.msg.startswith(
            "NOTICE: This code uses __future__ functionality in the ISPyB API."
        ):
            return 0

        return 1


# --------------------------------------------------------------------------------
@pytest.fixture(scope="function")
def output_directory(request):
    # TODO: Better way to get a newline in conftest after pytest emits the test class name.
    print("")

    # Tmp directory which we can write into.
    output_directory = "/tmp/%s/%s/%s" % (
        "/".join(__file__.split("/")[-3:-1]),
        request.cls.__name__,
        request.function.__name__,
    )

    # Tmp directory which we can write into.
    if os.path.exists(output_directory):
        shutil.rmtree(output_directory, ignore_errors=False, onerror=None)
    os.makedirs(output_directory)

    # logger.debug("output_directory is %s" % (output_directory))

    yield output_directory



# ----------------------------------------------------------------------------------------
class Tablenames:
    HEAD = "soakdb"
    BODY = "mainTable"
    VISIT = "visit"


# ----------------------------------------------------------------------------------------
class HeadFieldnames:
    Version = "Version"
    LabVisit = "LabVisit"
    Path = "Path"
    Protein = "Protein"
    DropVolume = "DropVolume"
    CrystalsPerBatch = "CrystalsPerBatch"
    OneBatchPerPlate = "OneBatchPerPlate"
    CompoundStock = "CompoundStock"
    SolventPercent = "SolventPercent"
    CryoStock = "CryoStock"
    DesiredCryo = "DesiredCryo"
    CryoLocation = "CryoLocation"
    DesiredSoakTime = "DesiredSoakTime"
    CrystalStartNumber = "CrystalStartNumber"
    BeamlineVisit = "BeamlineVisit"
    SpaceGroup = "SpaceGroup"
    A = "A"
    B = "B"
    C = "C"
    alpha = "alpha"
    beta = "beta"
    gamma = "gamma"
    Recipe = "Recipe"
    Resolution = "Resolution"
    CentringMethod = "CentringMethod"
    EUBOpen = "EUBOpen"
    iNEXT = "iNEXT"
    Covid19 = "Covid19"
    ILOXchem = "ILOXchem"


# ----------------------------------------------------------------------------------------
class BodyFieldnames:
    ID = "ID"
    LabVisit = "LabVisit"
    LibraryPlate = "LibraryPlate"
    SourceWell = "SourceWell"
    LibraryName = "LibraryName"
    CompoundSMILES = "CompoundSMILES"
    CompoundCode = "CompoundCode"
    CrystalPlate = "CrystalPlate"
    CrystalWell = "CrystalWell"
    EchoX = "EchoX"
    EchoY = "EchoY"
    DropVolume = "DropVolume"
    ProteinName = "ProteinName"
    BatchNumber = "BatchNumber"
    CompoundStockConcentration = "CompoundStockConcentration"
    CompoundConcentration = "CompoundConcentration"
    SolventFraction = "SolventFraction"
    SoakTransferVol = "SoakTransferVol"
    SoakStatus = "SoakStatus"
    SoakTimestamp = "SoakTimestamp"
    CryoStockFraction = "CryoStockFraction"
    CryoFraction = "CryoFraction"
    CryoWell = "CryoWell"
    CryoTransferVolume = "CryoTransferVolume"
    CryoStatus = "CryoStatus"
    CryoTimestamp = "CryoTimestamp"
    SoakingTime = "SoakingTime"
    HarvestStatus = "HarvestStatus"
    CrystalName = "CrystalName"
    Puck = "Puck"
    PuckPosition = "PuckPosition"
    PinBarcode = "PinBarcode"
    MountingResult = "MountingResult"
    MountingArrivalTime = "MountingArrivalTime"
    MountedTimestamp = "MountedTimestamp"
    MountingTime = "MountingTime"
    ispybStatus = "ispybStatus"
    DataCollectionVisit = "DataCollectionVisit"
    SoakDBComments = "SoakDBComments"


# ----------------------------------------------------------------------------------------
class VisitFieldnames:
    CREATED_ON = "created_on"
    VISITID = "visitid"



# Use standard logging in this module.
import logging

# Exceptions.
from soakdb3_api.exceptions import NotFound

# Class managing list of things.
from soakdb3_api.things import Things

logger = logging.getLogger(__name__)


class Databases(Things):
    """
    List of available databases.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification):
        """"""

        database_class = self.lookup_class(specification["type"])

        try:
            database_object = database_class(specification)
        except Exception as exception:
            raise RuntimeError(
                "unable to build database object for type %s" % (database_class)
            ) from exception

        return database_object

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        if class_type == "soakdb3_lib.databases.aiosqlite":
            from soakdb3_lib.databases.aiosqlite import Aiosqlite

            return Aiosqlite

        raise NotFound("unable to get database class for type %s" % (class_type))



import asyncio
import logging
import multiprocessing
import os

import pytest
import yaml

# Utilities.
from dls_utilpack.describe import describe

# Exceptions.
from soakdb3_api.exceptions import NotFound

# Configurator.
from soakdb3_lib.configurators.configurators import Configurators

# Environment variables with some extra functionality.
from soakdb3_lib.envvar import Envvar

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestConfiguratorDirectGood:
    def test(self, constants, logging_setup, output_directory):
        """ """

        # This is the configuration which the ConfiguratorTester will check.
        configuration = {
            "output_directory": "${output_directory}",
            "multi": {"level1": {"level2": "value_at_level2"}},
            "key_to_remove": "value_to_remove",
        }

        # Write the yaml file target to load.
        yaml_filename = f"{output_directory}/configuration.yml"
        with open(yaml_filename, "w") as yaml_stream:
            yaml.dump(
                configuration, yaml_stream, default_flow_style=False, sort_keys=False
            )

        # This is the object type we want to run the test on.
        specification = {
            "type": "soakdb3_lib.configurators.yaml",
            "type_specific_tbd": {"filename": yaml_filename},
        }

        # Run the test in a coroutine.
        GoodConfiguratorDirectTester().main(constants, specification, output_directory)


# ----------------------------------------------------------------------------------------
class TestConfiguratorDirectBad:
    def test(self, constants, logging_setup, output_directory):
        """ """

        # Run the bad tests in a coroutine.
        BadConfiguratorDirectTester("RuntimeError", "does not contain keyword").main(
            constants, {}, output_directory
        )
        BadConfiguratorDirectTester("RuntimeError", "unable to instantiate").main(
            constants,
            {"type": "soakdb3_lib.configurators.yaml"},
            output_directory,
        )

        # Another yaml file.
        bad_yaml_filename = f"{output_directory}/configuration.bad.yml"
        specification = {
            "type": "soakdb3_lib.configurators.yaml",
            "type_specific_tbd": {"filename": bad_yaml_filename},
        }

        BadConfiguratorDirectTester("RuntimeError", "unable to read").main(
            constants,
            specification,
            output_directory,
        )

        with open(bad_yaml_filename, "w") as yaml_stream:
            yaml_stream.write("a: - some non-yaml stuff")

        BadConfiguratorDirectTester("RuntimeError", "unable to parse").main(
            constants,
            specification,
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class TestGoodConfigurationEnv:
    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_filename = f"{output_directory}/good_config1.yaml"
        with open(configuration_filename, "w") as stream:
            stream.write("")

        os.makedirs(f"{output_directory}/good_data_dir1")

        # ---------------------------------------------------
        environ = {
            Envvar.SOAKDB3_CONFIGFILE: configuration_filename,
        }
        GoodConfigurationEnvTester().main(constants, environ, output_directory)


# ----------------------------------------------------------------------------------------
class TestBadConfigurationEnv:
    def test(self, constants, logging_setup, output_directory):
        """ """

        # Run the bad tests in a coroutine.

        BadConfigurationEnvTester("RuntimeError", "SOAKDB3_CONFIGFILE is not set").main(
            constants,
            {},
            output_directory,
        )

        BadConfigurationEnvTester("RuntimeError", "does_not_exist.yaml").main(
            constants,
            {
                Envvar.SOAKDB3_CONFIGFILE: "does_not_exist.yaml",
            },
            output_directory,
        )


# ----------------------------------------------------------------------------------------
class Base:
    """
    This is a base class for tests which use XchemBeContext.
    """

    def main(self, constants, specification, output_directory):
        """
        This is the main program which calls the test using asyncio.
        """

        multiprocessing.current_process().name = "main"

        failure_message = None
        try:
            # Run main test in asyncio event loop.
            asyncio.run(
                self._main_coroutine(constants, specification, output_directory)
            )

        except Exception as exception:
            logger.exception(
                "unexpected exception in the test method", exc_info=exception
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)


# ----------------------------------------------------------------------------------------
class GoodConfiguratorDirectTester(Base):
    """
    Class to test any type of configurator.
    """

    async def _main_coroutine(self, constants, specification, output_directory):
        """ """

        # Build the object of the type we want to test.
        configurator = Configurators().build_object(specification)

        # Perform symbol replacement.
        configurator.substitute({"output_directory": output_directory})

        # Remove top level key of this name.
        configurator.remove(["key_to_remove"])

        # Get the current dict representation of the configuration.
        current = await configurator.load()

        # Check the symbol has been replaced.
        assert current["output_directory"] == output_directory

        # Check the key has been removed.
        assert "key_to_remove" not in current

        # Extract multilevel value.
        value = configurator.require("multi.level1.level2")
        assert value == "value_at_level2"

        # Check a good error on non-existence of key.
        with pytest.raises(NotFound):
            value = configurator.require("multi.level1.level3")

        # Check a good error on non-existence of key.
        with pytest.raises(NotFound):
            value = configurator.require("multi.level1.level2.level3")

        # Substitute in token.
        value = configurator.resolve("${output_directory}/${multi.level1.level2}")
        assert value == f"{output_directory}/value_at_level2"


# ----------------------------------------------------------------------------------------
class BadConfiguratorDirectTester(Base):
    """
    Class to test any type of configurator.
    """

    def __init__(self, expected_exception_type, expected_message_snippet):
        self.__expected_exception_type = expected_exception_type
        self.__expected_message_snippet = expected_message_snippet

    async def _main_coroutine(self, constants, specification, output_directory):
        """ """

        got_exception = None
        try:
            # Build the object of the type we want to test.
            configurator = Configurators().build_object(specification)

            # Load the configuration fromt the source provider.
            await configurator.load()

            logger.info(describe("loaded yaml", configurator.get_current()))
        except Exception as exception:
            got_exception = exception

        got_exception_type = type(got_exception).__name__

        assert got_exception_type == self.__expected_exception_type
        assert self.__expected_message_snippet in str(got_exception)


# ----------------------------------------------------------------------------------------
class GoodConfigurationEnvTester(Base):
    """
    Class to test any type of configurator.
    """

    async def _main_coroutine(self, constants, environ, output_directory):
        """ """

        # Define the configuration source.
        Configurators().build_object_from_environment(environ=environ)


# ----------------------------------------------------------------------------------------
class BadConfigurationEnvTester(Base):
    """
    Class to test any type of configurator.
    """

    def __init__(self, expected_exception_type, expected_message_snippet):
        self.__expected_exception_type = expected_exception_type
        self.__expected_message_snippet = expected_message_snippet

    async def _main_coroutine(self, constants, environ, output_directory):
        """ """

        got_exception = None
        try:

            # Define the configuration source.
            Configurators().build_object_from_environment(environ=environ)

        except Exception as exception:
            # logger.error("expected error", exc_info=exception)
            got_exception = exception

        got_exception_type = type(got_exception).__name__

        assert got_exception_type == self.__expected_exception_type
        assert self.__expected_message_snippet in str(got_exception)



import json
from argparse import ArgumentParser

from soakdb3_lib.version import meta, version


def get_parser():
    parser = ArgumentParser(
        description="Command line accompanying the soakdb3 library."
    )
    parser.add_argument(
        "--version",
        action="version",
        version=version(),
        help="Print version string.",
    )
    parser.add_argument(
        "--version-json",
        action="store_true",
        help="Print version stack in json.",
    )
    return parser


def main(args=None):
    parser = get_parser()
    args = parser.parse_args(args)

    if args.version_json:
        print(json.dumps(meta(), indent=4))


if __name__ == "__main__":
    main()



import copy
import csv
import logging
import os
import re

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.require import require

# Database constants.
from soakdb3_api.databases.constants import Tablenames

# Base class for generic things.
from soakdb3_api.thing import Thing

# Database manager.
from soakdb3_lib.databases.databases import Databases

# Version for health response.
from soakdb3_lib.version import version as soakdb3_lib_version

logger = logging.getLogger(__name__)

thing_type = "soakdb3_lib.datafaces.aiosqlite"


class Aiosqlite(Thing):
    """
    Implementation of dataface on top of underlying database API provider presumed to be sqlite.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        Thing.__init__(self, thing_type, specification)

        self.__type_specific = require(
            f"{callsign(self)} specification",
            self.specification(),
            "type_specific_tbd",
        )

        self.__dbspec = require(
            f"{callsign(self)} specification",
            self.__type_specific,
            "database",
        )

        self.__visitid_mappings = require(
            f"{callsign(self)} specification",
            self.__type_specific,
            "visitid_mappings",
        )

        # Cache of database objects we have created.
        self.__cache = {}
        self.__cache_cvs_directories = {}

    # ----------------------------------------------------------------------------------------
    async def start(self):
        # There is no visitid known at start of the server.
        # They are built up as requested by clients.

        pass

    # ----------------------------------------------------------------------------------------
    async def disconnect(self):
        visitids = list(self.__cache.keys())
        for visitid in visitids:
            database = self.__cache.pop(visitid)
            await database.disconnect()
            database = None

    # ----------------------------------------------------------------------------------------
    async def establish_database_connection(self, visitid):

        database = self.__cache.get(visitid)
        if database is None:

            # Get path from visitid.
            visitid_path = self.__visitid2path(visitid)

            logger.debug(f"visitid {visitid} is for path {visitid_path}")

            if not os.path.isdir(visitid_path):
                raise RuntimeError(f"{visitid_path} is not a directory")

            # Get copy of specification template.
            dbspec = copy.deepcopy(self.__dbspec)

            dbspec["filename"] = f"{visitid_path}/database/soakDBDataFile.sqlite"
            dbspec["backup_directory"] = f"{visitid_path}/database/DataFileBackups"

            self.__cache_cvs_directories[visitid] = f"{visitid_path}/lab36"

            database = Databases().build_object(dbspec)
            await database.connect()

            # Cache database connections by visitid.
            self.__cache[visitid] = database

        return database

    # ----------------------------------------------------------------------------------------
    def __visitid2path(self, visitid: str) -> str:
        """
        Give back a file path given the visitid.

        Args:
            visitid (str): visitid
                At the current time, excel gives the visitid as windows-type path with leading drive letter.

        Returns:
            str: Linux filesystem path root of the database, backup and csv files.

        """

        # Go through all the mappings defined in the specification.
        for visitid_mapping in self.__visitid_mappings:
            # This mapping calls for regex replacement?
            if visitid_mapping["action"] == "regex_replace":
                pattern = visitid_mapping["pattern"]
                replace = visitid_mapping["replace"]

                # The visitid matches this mapping?
                match = re.search(pattern, visitid)
                if match is not None:
                    # Replace with the substitution.
                    path = re.sub(pattern, replace, visitid)
                    return path

        return visitid

    # # ----------------------------------------------------------------------------------------
    # async def reinstance(self):
    #     """"""

    #     # TODO: Consider if reinstance is a necessary method.
    #     if self.__database is None:
    #         return

    #     self.__database = self.__database.reinstance()

    # ----------------------------------------------------------------------------------------
    async def backup(self, visitid):
        """"""
        database = await self.establish_database_connection(visitid)

        return await database.backup()

    # ----------------------------------------------------------------------------------------
    async def restore(self, visitid, nth):
        """"""
        database = await self.establish_database_connection(visitid)

        return await database.restore(nth)

    # ----------------------------------------------------------------------------------------
    async def query_for_dictionary(self, visitid, sql, subs=None, why=None):
        """"""
        database = await self.establish_database_connection(visitid)

        records = await database.query(sql, subs=subs, why=why)

        return records

    # ----------------------------------------------------------------------------------------
    async def query(self, visitid, sql, subs=None, why=None):
        """"""
        database = await self.establish_database_connection(visitid)

        records = await database.query(sql, subs=subs, why=why)

        # Make a VBA collection with the first record giving the field names.
        collection = []
        for index, record in enumerate(records):
            if index == 0:
                collection.append(list(record.keys()))
            collection.append(list(record.values()))

        return collection

    # ----------------------------------------------------------------------------------------
    async def execute(self, visitid, sql, subs=None, why=None):
        """"""
        database = await self.establish_database_connection(visitid)

        return await database.execute(sql, subs=subs, why=why)

    # ----------------------------------------------------------------------------------------
    async def insert(self, visitid, table_name, records, subs=None, why=None):
        """"""
        database = await self.establish_database_connection(visitid)

        if why is None:
            why = f"insert {len(records)} {table_name} records"

        await database.insert(table_name, records)

    # ----------------------------------------------------------------------------------------
    async def update(self, visitid, table_name, record, where, subs=None, why=None):
        """"""
        database = await self.establish_database_connection(visitid)

        if why is None:
            why = f"update {table_name} record"

        await database.update(table_name, record, where, why=why)

    # ----------------------------------------------------------------------------------------
    async def update_body_fields(self, visitid, fields):
        """
        Handle update request from a soakdb range change.
        """
        database = await self.establish_database_connection(visitid)

        table_name = Tablenames.BODY

        maximum_id = None
        # Group the updates to rows by their id.
        rows = {}
        new_ids = {}
        for field in fields:
            id = str(field["id"])
            # This is a new row created from within the client?
            if id[0] == "-":
                # We haven't assigned this record to a real database id yet?
                if id not in new_ids:
                    # We don't know the maximum id in the database yet?
                    if maximum_id is None:
                        maximum_id = await self.query_for_dictionary(
                            visitid, f"SELECT MAX(ID) AS MAX_ID FROM {table_name}"
                        )
                        maximum_id = maximum_id[0]["MAX_ID"]
                    if maximum_id is None:
                        maximum_id = 0
                    maximum_id += 1
                    new_ids[id] = str(maximum_id)
                # Translate the negative id to a new id in the database.
                id = new_ids[id]

            row = rows.get(id)
            if row is None:
                row = []
                rows[id] = row

            # VBA doesn't put a Nothing value in the dictionary, so treat missing value as None.
            value = field.get("value")
            row.append({"field": field["field"], "value": value})

        # Get list of existing ids for those we want to update.
        id_csv = ", ".join(rows.keys())
        existing_rows = await self.query_for_dictionary(
            visitid, f"SELECT ID FROM {table_name} WHERE ID IN ({id_csv})"
        )
        existing_ids = []
        for existing_row in existing_rows:
            existing_ids.append(str(existing_row["ID"]))

        new_rows = []
        for updated_id in rows.keys():
            if updated_id not in existing_ids:
                new_rows.append({"ID": updated_id})

        logger.debug(
            f"inserting {len(new_rows)} new rows"
            f" and updating {len(rows.keys())} rows"
        )

        try:
            # Do the inserts at the start of the transaction.
            await database.insert(
                Tablenames.BODY,
                new_rows,
                should_commit=False,
            )

            # Perform the updates on each row.
            for id, row in rows.items():
                sets = []
                subs = []
                for k in row:
                    field = k["field"]
                    value = k["value"]
                    sets.append(f"`{field}` = ?")
                    subs.append(value)
                subs.append(id)

                sets = ", ".join(sets)
                sql = f"UPDATE {table_name} SET {sets} WHERE ID = ?"

                # Add the update to the transaction.
                await database.execute(
                    sql,
                    subs=subs,
                    should_commit=False,
                )
        except Exception:
            # Don't keep any partial operation.
            database.rollback()
            raise

        # Commit all the updates.
        logger.debug(
            f"committing {len(new_rows)} new rows"
            f" and updating {len(rows.keys())} rows"
        )
        await database.commit()
        logger.debug("committed")

    # ----------------------------------------------------------------------------------------
    async def update_head_fields(self, visitid, fields):
        """
        Handle update request from a soakdb range change.
        """

        table_name = Tablenames.HEAD

        # Perform the updates on the only row row.
        sets = []
        subs = []
        for k in fields:
            field = k["field"]
            value = k["value"]
            sets.append(f"`{field}` = ?")
            subs.append(value)

        sets = ", ".join(sets)
        sql = f"UPDATE {table_name} SET {sets}"

        await self.execute(visitid, sql, subs=subs, why="update head table")

    # ----------------------------------------------------------------------------------------
    async def write_csv(self, visitid, rows, filename):
        """
        Handle request to write rows as csv file.

        The rows are assumed fully prepared ahead of time.

        The filename's path should be a subdirectory (no leading slash).
        """

        # We don't really need the database itself, since all the data rows are provided as arguments.
        # However, this sets up the csv_directory for this visit.
        await self.establish_database_connection(visitid)

        # Get the csv directory from the specification.
        csv_directory = self.__cache_cvs_directories.get(visitid)
        if csv_directory is None:
            raise RuntimeError(
                f"no database connection has yet been made for visitid {visitid}"
            )

        filename = f"{csv_directory}/{filename}"
        await self.__create_directory(filename)

        with open(filename, "w") as stream:

            # Create the csv writer.
            writer = csv.writer(stream)

            # Write all rows to the csv file.
            writer.writerows(rows)

    # ----------------------------------------------------------------------------------------
    async def __create_directory(self, filename):

        directory, filename = os.path.split(filename)

        if not os.path.exists(directory):
            # Make sure that parent directories which get created will have public permission.
            umask = os.umask(0)
            os.umask(umask & ~0o0777)
            os.makedirs(directory)
            os.umask(umask)

    # ----------------------------------------------------------------------------------------
    async def report_health(self):
        """"""

        report = {}

        report["alive"] = True
        report["version"] = soakdb3_lib_version()

        return report



# Use standard logging in this module.
import logging
import os
from typing import Optional

from dls_servbase_lib.configurators.configurators import (
    dls_servbase_configurators_set_default,
)

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.require import require

# Exceptions.
from soakdb3_api.exceptions import NotFound

# Class managing list of things.
from soakdb3_api.things import Things

# Environment variables with some extra functionality.
from soakdb3_lib.envvar import Envvar

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------------------
__default_configurator = None


def configurators_set_default(configurator):
    global __default_configurator
    __default_configurator = configurator

    # We share configurator with servbase.
    dls_servbase_configurators_set_default(configurator)


def configurators_get_default():
    global __default_configurator
    if __default_configurator is None:
        raise RuntimeError("configurators_get_default instance is None")
    return __default_configurator


def configurators_has_default():
    global __default_configurator
    return __default_configurator is not None


# -----------------------------------------------------------------------------------------


class Configurators(Things):
    """
    Configuration loader.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification):
        """"""

        configurator_class = self.lookup_class(
            require(f"{callsign(self)} specification", specification, "type")
        )

        try:
            configurator_object = configurator_class(specification)
        except Exception as exception:
            raise RuntimeError(
                "unable to instantiate configurator object from module %s"
                % (configurator_class.__module__)
            ) from exception

        return configurator_object

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        if class_type == "soakdb3_lib.configurators.yaml":
            from soakdb3_lib.configurators.yaml import Yaml

            return Yaml

        raise NotFound("unable to get configurator class for type %s" % (class_type))

    # ----------------------------------------------------------------------------------------
    def build_object_from_environment(
        self,
        environ: Optional[dict] = None,
        args_dict: Optional[dict] = None,
    ):

        configuration_keyword = "configuration"

        configurator_filename = None

        if args_dict is not None:
            configurator_filename = args_dict.get(configuration_keyword)

        if configurator_filename is not None:
            # Make sure the path exists.
            if not os.path.exists(configurator_filename):
                raise RuntimeError(
                    f"unable to find --{configuration_keyword} file {configurator_filename}"
                )
        else:
            # Get the explicit name of the config file.
            bxflow_configfile = Envvar(
                Envvar.SOAKDB3_CONFIGFILE,
                environ=environ,
            )

            # Config file is explicitly named?
            if bxflow_configfile.is_set:
                # Make sure the path exists.
                configurator_filename = bxflow_configfile.value
                if not os.path.exists(configurator_filename):
                    raise RuntimeError(
                        f"unable to find {Envvar.SOAKDB3_CONFIGFILE} {configurator_filename}"
                    )
            # Config file is not explicitly named?
            else:
                raise RuntimeError(
                    f"command line --{configuration_keyword} not given"
                    f" and environment variable {Envvar.SOAKDB3_CONFIGFILE} is not set"
                )
        configurator = self.build_object(
            {
                "type": "soakdb3_lib.configurators.yaml",
                "type_specific_tbd": {"filename": configurator_filename},
            }
        )

        configurator.substitute(
            {"configurator_directory": os.path.dirname(configurator_filename)}
        )

        return configurator



# Use standard logging in this module.
import logging

import yaml

# Exceptions.
from soakdb3_api.exceptions import NotFound

# Class managing list of things.
from soakdb3_api.things import Things

logger = logging.getLogger(__name__)

# -----------------------------------------------------------------------------------------


class Contexts(Things):
    """
    Context loader.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, name=None):
        Things.__init__(self, name)

    # ----------------------------------------------------------------------------------------
    def build_object(self, specification):
        """"""

        if not isinstance(specification, dict):
            with open(specification, "r") as yaml_stream:
                specification = yaml.safe_load(yaml_stream)

        xchem_be_context_class = self.lookup_class(specification["type"])

        try:
            xchem_be_context_object = xchem_be_context_class(specification)
        except Exception as exception:
            raise RuntimeError(
                "unable to build xchem_be_context object for type %s"
                % (xchem_be_context_class)
            ) from exception

        return xchem_be_context_object

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        if class_type == "soakdb3_lib.contexts.classic":
            from soakdb3_lib.contexts.classic import Classic

            return Classic

        raise NotFound(
            "unable to get xchem_be_context class for type %s" % (class_type)
        )



import logging
import os

logger = logging.getLogger(__name__)


class Envvar:
    """Class which covers environment variables, with default values."""

    SOAKDB3_CONFIGFILE = "SOAKDB3_CONFIGFILE"
    XCHEM_BEFLOW_DATA = "XCHEM_BEFLOW_DATA"
    XCHEM_BEFLOW_DLS_ROOT = "XCHEM_BEFLOW_DLS_ROOT"
    BEAMLINE = "BEAMLINE"
    VISIT_YEAR = "VISIT_YEAR"
    VISIT = "VISIT"

    def __init__(self, name, **kwargs):

        environ = kwargs.get("environ")

        if environ is None:
            environ = os.environ

        self.name = name
        self.is_set = False
        self.value = None

        if name in environ:
            self.is_set = True
            self.value = environ[name]
        else:
            if "default" in kwargs:
                self.is_set = True
                self.value = kwargs["default"]
            else:
                self.is_set = False
                self.value = None



import logging
import os

# Configurator.
from soakdb3_lib.configurators.configurators import (
    Configurators,
    configurators_set_default,
)


class ArgKeywords:
    CONFIGURATION = "configuration"


logger = logging.getLogger(__name__)


class Base:
    """
    Base class for subcommands.  Handles details like configuration.
    """

    def __init__(self, args):
        self._args = args

        self.__temporary_directory = None

    # ----------------------------------------------------------------------------------------
    def get_configurator(self, args_dict: dict):

        configurator = Configurators().build_object_from_environment(
            args_dict=args_dict
        )

        substitutions = {
            "CWD": os.getcwd(),
            "HOME": os.environ.get("HOME", "HOME"),
            "PYTHONPATH": os.environ.get("PYTHONPATH", "PYTHONPATH"),
            "USER": os.environ.get("USER", "USER"),
        }

        configurator.substitute(substitutions)

        # Set this as the default configurator so it is available everywhere.
        configurators_set_default(configurator)

        return configurator



import csv
import logging

# Object managing datafaces.
from soakdb3_api.datafaces.datafaces import datafaces_get_default

# Context creator.
from soakdb3_lib.contexts.contexts import Contexts

# Base class for the tester.
from tests.base_context_tester import BaseContextTester

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestWriteCsv:
    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/laptop.yaml"
        WriteCsvTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class WriteCsvTester(BaseContextTester):
    """
    Class to test the write_csv API call through the server.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        configurator = self.get_configurator()

        context_configuration = await configurator.load()
        context = Contexts().build_object(context_configuration)

        # The visitid to match what is in the test configuration yaml.
        visitid = output_directory

        async with context:
            dataface = datafaces_get_default()

            data = [
                ["a", "1"],
                ["b", "2"],
                ["c", '"3"'],
                ["d", "4,5,6"],
                ["e", "line1\nline2"],
                ["f", ""],
                ["g"],
                ["h", "done"],
            ]

            filename = "echo/Batchfile.csv"

            # Write csv.
            await dataface.write_csv(visitid, data, filename)

            # Read csv.
            fieldnames = [0, 1]
            restval = "none"
            with open(f"{output_directory}/lab36/{filename}") as stream:
                reader = csv.DictReader(stream, fieldnames=fieldnames, restval=restval)
                i = 0
                for row in reader:
                    for col, value in row.items():
                        # Data row has at least this many columns?
                        if col < len(data[i]):
                            assert value == data[i][col], f"line {i} column {col}"
                        else:
                            assert value == restval, f"line {i} column {col}"
                    i = i + 1



import logging

from dls_servbase_lib.base_aiohttp import BaseAiohttp as DlsServbaseBaseAiohttp

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------
class BaseAiohttp(DlsServbaseBaseAiohttp):
    """
    Object representing a a process which receives requests from aiohttp.
    """

    pass



import asyncio
import logging
import multiprocessing
import os

import pytest

# Configurator.
from soakdb3_lib.configurators.configurators import (
    Configurators,
    configurators_set_default,
)

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class BaseContextTester:
    """
    This is a base class for tests which use XchemBeContext.
    """

    def __init__(self):
        # For a capturing the news events as they arrive.
        self.consumed_news = []
        self.tasks_execution_outputs = {}
        self.residuals = ["stdout.txt", "stderr.txt", "main.log"]

    def main(self, constants, configuration_file, output_directory):
        """
        This is the main program which calls the test using asyncio.
        """

        # Save these for when the configuration is loaded.
        self.__configuration_file = configuration_file
        self.__output_directory = output_directory

        multiprocessing.current_process().name = "main"

        # self.__blocked_event = asyncio.Event()

        failure_message = None
        try:
            # Run main test in asyncio event loop.
            asyncio.run(self._main_coroutine(constants, output_directory))

        except Exception as exception:
            logger.exception(
                "unexpected exception in the test method", exc_info=exception
            )
            failure_message = str(exception)

        if failure_message is not None:
            pytest.fail(failure_message)

    # ----------------------------------------------------------------------------------------
    def get_configurator(self):

        configurator = Configurators().build_object(
            {
                "type": "soakdb3_lib.configurators.yaml",
                "type_specific_tbd": {"filename": self.__configuration_file},
            }
        )

        # For convenience, always do these replacement.
        configurator.substitute({"output_directory": self.__output_directory})

        # Add various things from the environment into the configurator.
        configurator.substitute(
            {
                "CWD": os.getcwd(),
                "PYTHONPATH": os.environ.get("PYTHONPATH", "PYTHONPATH"),
            }
        )

        # Set the global value of our configurator which might be used in other modules.
        configurators_set_default(configurator)

        return configurator



import logging

# Things created in the context.
from soakdb3_api.datafaces.datafaces import Datafaces, datafaces_set_default

logger = logging.getLogger(__name__)


class Context:
    """
    Client context for a dataface object.
    On entering, it creates the object according to the specification (a dict).
    On exiting, it closes client connection.

    The aenter and aexit methods are exposed for use by an enclosing context.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        self.__specification = specification
        self.__dataface = None

    # ----------------------------------------------------------------------------------------
    async def __aenter__(self):
        """ """

        await self.aenter()

    # ----------------------------------------------------------------------------------------
    async def __aexit__(self, type, value, traceback):
        """ """

        await self.aexit()

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        # Build the object according to the specification.
        self.__dataface = Datafaces().build_object(self.__specification)

        # If there is more than one dataface, the last one defined will be the default.
        datafaces_set_default(self.__dataface)

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        if self.__dataface is not None:
            await self.__dataface.close_client_session()

            # Clear the global variable.  Important between pytests.
            datafaces_set_default(None)



import argparse
import json

import dls_mainiac_lib.version

from soakdb3_lib import __version__


# ----------------------------------------------------------
def version():
    """
    Current version.
    """

    return __version__


# ----------------------------------------------------------
def meta(given_meta=None):
    """
    Returns version information as a dict.
    Adds version information to given meta, if any.
    """
    s = {}
    s["soakdb3_cli"] = version()
    s.update(dls_mainiac_lib.version.meta())

    if given_meta is not None:
        given_meta.update(s)
    else:
        given_meta = s
    return given_meta


# ----------------------------------------------------------
def main():

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--json",
        action="store_true",
        help="Print version stack in json.",
    )

    # -------------------------------------------------------------------------
    given_args, remaining_args = parser.parse_known_args()

    if given_args.json:
        print(json.dumps(meta(), indent=4))
    else:
        print(version())


# ----------------------------------------------------------
if __name__ == "__main__":
    main()



import logging

# Base class for all aiosqlite database objects.
from soakdb3_lib.databases.table_definitions import BodyTable, HeadTable, VisitTable

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class DatabaseDefinition:
    """
    Class which defines the database tables and revision migration path.
    Used in concert with the normsql class.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self):
        """
        Construct object.  Do not connect to database.
        """

        self.LATEST_REVISION = 1

    # ----------------------------------------------------------------------------------------
    async def add_table_definitions(self):
        """
        Make all the table definitions.
        """

        # Table schemas in our database.
        self.add_table_definition(HeadTable())
        self.add_table_definition(BodyTable())
        self.add_table_definition(VisitTable())



import logging

# Class for an aiohttp client.
from soakdb3_api.aiohttp_client import AiohttpClient

# XchemBeDataface protocolj things.
from soakdb3_api.datafaces.constants import Commands, Keywords

logger = logging.getLogger(__name__)


# ------------------------------------------------------------------------------------------
class Aiohttp:
    """
    Object implementing client side API for talking to the dataface server.
    Please see doctopic [A01].
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        self.__specification = specification

        self.__aiohttp_client = AiohttpClient(
            specification["type_specific_tbd"]["aiohttp_specification"],
        )

    # ----------------------------------------------------------------------------------------
    def specification(self):
        return self.__specification

    # ----------------------------------------------------------------------------------------
    async def backup(self, visitid):
        """"""
        return await self.__send_protocolj("backup", visitid)

    # ----------------------------------------------------------------------------------------
    async def restore(self, visitid, nth):
        """"""
        return await self.__send_protocolj("restore", visitid, nth)

    # ----------------------------------------------------------------------------------------
    async def query(self, visitid, sql, subs=None, why=None):
        """"""
        return await self.__send_protocolj("query", visitid, sql, subs=subs, why=why)

    # ----------------------------------------------------------------------------------------
    async def query_for_dictionary(self, visitid, sql, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "query_for_dictionary", visitid, sql, subs=subs, why=why
        )

    # ----------------------------------------------------------------------------------------
    async def execute(self, visitid, sql, subs=None, why=None):
        """"""
        return await self.__send_protocolj("execute", visitid, sql, subs=subs, why=why)

    # ----------------------------------------------------------------------------------------
    async def insert(self, visitid, table_name, records, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "insert", visitid, table_name, records, subs=subs, why=why
        )

    # ----------------------------------------------------------------------------------------
    async def update(self, visitid, table_name, record, where, subs=None, why=None):
        """"""
        return await self.__send_protocolj(
            "update", visitid, table_name, record, where, subs=subs, why=why
        )

    # ----------------------------------------------------------------------------------------
    async def update_head_fields(self, visitid, fields):
        """
        Handle update request from a soakdb range change.
        """

        return await self.__send_protocolj("update_head_fields", visitid, fields)

    # ----------------------------------------------------------------------------------------
    async def update_body_fields(self, visitid, fields):
        """
        Handle update request from a soakdb range change.
        """

        return await self.__send_protocolj("update_body_fields", visitid, fields)

    # ----------------------------------------------------------------------------------------
    async def write_csv(self, visitid, rows, filename):
        """
        Handle request to write rows as csv file.

        The rows are assumed fully prepared ahead of time.

        The filename's path should be a subdirectory (no leading slash).

        """

        return await self.__send_protocolj("write_csv", visitid, rows, filename)

    # ----------------------------------------------------------------------------------------
    async def report_health(self):
        """"""
        return await self.__send_protocolj("report_health")

    # ----------------------------------------------------------------------------------------
    async def __send_protocolj(self, function, *args, **kwargs):
        """"""

        return await self.__aiohttp_client.client_protocolj(
            {
                Keywords.COMMAND: Commands.EXECUTE,
                Keywords.PAYLOAD: {
                    "function": function,
                    "args": args,
                    "kwargs": kwargs,
                },
            },
        )

    # ----------------------------------------------------------------------------------------
    async def close_client_session(self):
        """"""

        if self.__aiohttp_client is not None:
            await self.__aiohttp_client.close_client_session()

    # ----------------------------------------------------------------------------------------
    async def client_report_health(self):
        """"""

        return await self.__aiohttp_client.client_report_health()



import logging

from dls_utilpack.callsign import callsign

# Utilities.
from dls_utilpack.explain import explain

# Base class which maps flask requests to methods.
from soakdb3_lib.contexts.base import Base

# Contexts.
from soakdb3_lib.datafaces.context import Context as XchemBeDatafaceContext

logger = logging.getLogger(__name__)


thing_type = "soakdb3_lib.contexts.classic"


class Classic(Base):
    """
    Object representing an event dataface connection.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        Base.__init__(self, thing_type, specification)

        self.__dataface = None

    # ----------------------------------------------------------------------------------------
    async def __aenter__(self):
        """ """
        logger.debug(f"entering {callsign(self)} context")

        try:

            try:
                specification = self.specification().get("dataface_specification")
                if specification is not None:
                    logger.debug(f"at entering position {callsign(self)} DATAFACE")
                    self.__dataface = XchemBeDatafaceContext(specification)
                    await self.__dataface.aenter()
            except Exception as exception:
                raise RuntimeError(
                    explain(exception, f"creating {callsign(self)} dataface context")
                )

        except Exception as exception:
            await self.aexit()
            raise RuntimeError(explain(exception, f"entering {callsign(self)} context"))

        logger.debug(f"entered {callsign(self)} context")

    # ----------------------------------------------------------------------------------------
    async def __aexit__(self, type, value, traceback):
        """ """

        await self.aexit()

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        logger.debug(f"exiting {callsign(self)} context")

        if self.__dataface is not None:
            logger.debug(f"at exiting position {callsign(self)} DATAFACE")
            try:
                await self.__dataface.aexit()
            except Exception as exception:
                logger.error(
                    explain(exception, f"exiting {callsign(self.__dataface)} context"),
                    exc_info=exception,
                )
            self.__dataface = None

        logger.debug(f"exited {callsign(self)} context")

    # ----------------------------------------------------------------------------------------
    async def __dead_or_alive(self, context, dead, alive):

        if context is not None:
            try:
                # A server was defined for this context?
                if await context.is_process_started():
                    if await context.is_process_alive():
                        alive.append(context)
                    else:
                        dead.append(context)
            except Exception:
                raise RuntimeError(
                    f"unable to determine dead or alive for context {callsign(context)}"
                )

    # ----------------------------------------------------------------------------------------
    async def __dead_or_alive_all(self):
        """
        Return two lists, one for dead and one for alive processes.
        TODO: Parallelize context process alive/dead checking.
        """

        dead = []
        alive = []

        await self.__dead_or_alive(self.__dataface, dead, alive)

        return dead, alive

    # ----------------------------------------------------------------------------------------
    async def is_any_process_alive(self):
        """
        Check all configured processes, return if any alive.
        """
        dead, alive = await self.__dead_or_alive_all()

        # logger.debug(f"[PIDAL] {len(dead)} processes are dead, {len(alive)} are alive")

        return len(alive) > 0

    # ----------------------------------------------------------------------------------------
    async def is_any_process_dead(self):
        """
        Check all configured processes, return if any alive.
        """
        dead, alive = await self.__dead_or_alive_all()

        return len(dead) > 0



"""
The version of the package can be returned as a single string or a dict.

When a string, it comes from the package __version__.
When a dict, it also has __version__,
as well as versions of other depdency packages.
"""

from typing import Optional

import dls_mainiac_lib.version
import dls_servbase_lib.version
import dls_utilpack.version

from soakdb3_lib import __version__


# ----------------------------------------------------------
def version() -> str:
    """
    Version of the soakdb3 package as a string.
    """

    return __version__


# ----------------------------------------------------------
def meta(given_meta: Optional[dict] = None) -> dict:
    """
    Returns version information from the soakdb3 package
    and its dependencies as a dict.
    Adds version information to a given meta dict if it was provided.
    """

    meta = {}
    meta["soakdb3_lib"] = version()
    meta.update(dls_servbase_lib.version.meta())
    meta.update(dls_mainiac_lib.version.meta())
    meta.update(dls_utilpack.version.meta())

    if given_meta is not None:
        given_meta.update(meta)
    else:
        given_meta = meta
    return given_meta



import logging

from soakdb3_api.datafaces.context import Context as ApiDatafaceContext

# Base class for an asyncio context
from soakdb3_lib.contexts.base import Base as ContextBase

# Things created in the context.
from soakdb3_lib.datafaces.datafaces import Datafaces

logger = logging.getLogger(__name__)


thing_type = "soakdb3_lib.datafaces.context"


class Context(ContextBase):
    """
    Asyncio context for a dataface server object.
    On entering, it creates the object according to the specification (a dict).
    If configured, it starts the server as a coroutine, thread or process.
    On exiting, it commands the server to shut down.

    The enter and exit methods are exposed for use during testing.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        ContextBase.__init__(self, thing_type, specification)

        self.__api_dataface_context = None

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        # Build the object according to the specification.
        self.server = Datafaces().build_object(self.specification())

        if self.context_specification.get("start_as") == "coro":
            await self.server.activate_coro()

        elif self.context_specification.get("start_as") == "thread":
            await self.server.start_thread()

        elif self.context_specification.get("start_as") == "process":
            await self.server.start_process()

        self.__api_dataface_context = ApiDatafaceContext(self.specification())
        await self.__api_dataface_context.aenter()

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        if self.server is not None:
            # Put in request to shutdown the server.
            await self.server.client_shutdown()

        if self.__api_dataface_context is not None:
            await self.__api_dataface_context.aexit()



import logging

import pytest

from soakdb3_api.databases.constants import BodyFieldnames, HeadFieldnames, Tablenames

# Object managing datafaces.
from soakdb3_api.datafaces.datafaces import datafaces_get_default

# Context creator.
from soakdb3_lib.contexts.contexts import Contexts

# Base class for the tester.
from tests.base_context_tester import BaseContextTester

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestDatafaceHead:
    def test_dataface_laptop(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/laptop.yaml"
        DatafaceHeadTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestDatafaceBody:
    def test(self, constants, logging_setup, output_directory):
        """ """

        configuration_file = "tests/configurations/laptop.yaml"
        DatafaceBodyTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class DatafaceHeadTester(BaseContextTester):
    """
    Class to test the news.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        configurator = self.get_configurator()

        context_configuration = await configurator.load()
        context = Contexts().build_object(context_configuration)

        # The visitid to match what is in the test configuration yaml.
        visitid = output_directory

        async with context:
            dataface = datafaces_get_default()

            # Write one record.
            await dataface.insert(
                visitid,
                Tablenames.HEAD,
                [
                    {
                        HeadFieldnames.LabVisit: "x",
                        HeadFieldnames.Protein: "y",
                    }
                ],
            )

            # --------------------------------------------------------------------
            # Update like XChem soakdb range changed.
            fields = [
                {"field": HeadFieldnames.LabVisit, "value": "a"},
                {"field": HeadFieldnames.Protein, "value": "b"},
                {"field": HeadFieldnames.DropVolume, "value": 1.23},
            ]

            await dataface.update_head_fields(visitid, fields)

            z3_sql = "SELECT * from soakDB"
            records = await dataface.query_for_dictionary(visitid, z3_sql)
            assert len(records) == 1
            assert records[0][HeadFieldnames.LabVisit] == "a"
            assert records[0][HeadFieldnames.Protein] == "b"
            assert records[0][HeadFieldnames.DropVolume] == 1.23

            # Query for invalid visit.
            with pytest.raises(RuntimeError) as exception_info:
                records = await dataface.query_for_dictionary("badvisit", z3_sql)
            assert "badvisit" in str(exception_info.value)


# ----------------------------------------------------------------------------------------
class DatafaceBodyTester(BaseContextTester):
    """
    Class to test the news.
    """

    async def _main_coroutine(self, constants, output_directory):
        """ """

        configurator = self.get_configurator()

        context_configuration = await configurator.load()
        xchem_be_context = Contexts().build_object(context_configuration)

        # The visitid to match what is in the test configuration yaml.
        visitid = output_directory

        uuid1 = 1000
        uuid2 = 2000
        uuid3 = 3000

        async with xchem_be_context:
            dataface = datafaces_get_default()

            # Write one record.
            await dataface.insert(
                visitid,
                Tablenames.BODY,
                [
                    {
                        BodyFieldnames.LabVisit: "x",
                        BodyFieldnames.ID: uuid1,
                    }
                ],
            )

            all_sql = f"SELECT * FROM {Tablenames.BODY} ORDER BY ID ASC"
            records = await dataface.query(visitid, all_sql)
            assert len(records) == 2, "first %s count" % (all_sql)

            # Write two more records.
            await dataface.insert(
                visitid,
                Tablenames.BODY,
                [
                    {
                        BodyFieldnames.LabVisit: "y",
                        BodyFieldnames.ID: uuid2,
                    },
                    {
                        BodyFieldnames.LabVisit: "z",
                        BodyFieldnames.ID: uuid3,
                    },
                ],
            )
            records = await dataface.query(visitid, all_sql)
            assert len(records) == 4, "second %s count" % (all_sql)

            # Update one record.
            await dataface.update(
                visitid,
                Tablenames.BODY,
                {BodyFieldnames.LabVisit: "z2"},
                f"ID = '{uuid3}'",
            )
            z2_sql = f"SELECT * FROM {Tablenames.BODY} WHERE {BodyFieldnames.LabVisit} = 'z2' ORDER BY ID ASC"
            records = await dataface.query(visitid, z2_sql)
            assert len(records) == 2, "third %s count" % z2_sql

            # --------------------------------------------------------------------
            # Update like XChem soakdb range changed.
            fields = [
                {"id": uuid1, "field": BodyFieldnames.CompoundCode, "value": 15},
                {"id": uuid2, "field": BodyFieldnames.CompoundCode, "value": 20},
                {"id": uuid3, "field": BodyFieldnames.CompoundCode, "value": 20},
            ]

            await dataface.update_body_fields(visitid, fields)

            z3_sql = f"SELECT ID, {BodyFieldnames.CompoundCode} FROM mainTable ORDER BY ID ASC"
            records = await dataface.query(visitid, z3_sql)
            assert len(records) == 4, "fourth %s count" % z3_sql
            assert records[1][1] == "15", "first compound"
            assert records[2][1] == "20", "second compound"
            assert records[3][1] == "20", "third compound"

            # # Update two records to DEAD.
            # await database.xchem_be_body_table.update(
            #     {BodyFieldnames.LabVisit: "u2"}, f"ID IN ({uuid1}, {uuid2})"
            # )
            # u2_sql = (
            #     f"SELECT * FROM mainTable WHERE {BodyFieldnames.LabVisit} = 'u2' ORDER BY ID ASC"
            # )
            # records = await database.query(u2_sql)
            # assert len(records) == 2, "%s count" % u2_sql



import json

# Use standard logging in this module.
import logging

# Method to import a class from a file.
from dls_utilpack.import_class import import_class

# Exceptions.
from soakdb3_api.exceptions import DuplicateUuidException, NotFound

logger = logging.getLogger(__name__)


# -----------------------------------------------------------------------------
class Things:
    """
    Class managing list of things.
    """

    def __init__(self, name):
        self._name = name
        self._list = []
        self._dict = {}

    # -----------------------------------------------------------------------------
    def name(self):
        return self._name

    # -----------------------------------------------------------------------------
    def list(self):
        return self._list

    # -----------------------------------------------------------------------------
    def clear(self):
        self._list = []
        self._dict = {}

    # -----------------------------------------------------------------------------
    def len(self):
        return len(self._list)

    # -----------------------------------------------------------------------------
    def add(self, things):
        if not hasattr(things, "__iter__"):
            things = [things]

        for thing in things:
            if thing.uuid() not in self._dict:
                self._list.append(thing)
                self._dict[thing.uuid()] = thing
            else:
                raise DuplicateUuidException(
                    "%s not adding duplicate %s uuid %s"
                    % (self._name, thing.thing_type(), thing.uuid())
                )

    # -----------------------------------------------------------------------------
    def find(self, key, trait_name=None):
        if trait_name is None:
            if key in self._dict:
                return self._dict[key]
            raise NotFound("%s list does not have uuid %s" % (self._name, key))
        else:
            for thing in self._list:
                if thing.trait(trait_name) == key:
                    return thing
            raise NotFound(
                "%s list does not have %s %s" % (self._name, trait_name, key)
            )

    # -----------------------------------------------------------------------------
    def has(self, uuid):
        return uuid in self._dict

    # -----------------------------------------------------------------------------
    # If a string, parse for json, yaml or whatever.
    def parse_specification(self, specification):
        if isinstance(specification, dict):
            return specification

        if isinstance(specification, str):
            return json.loads(specification)

        raise RuntimeError(
            "specification is a %s but needs to be a dict or a string"
            % (type(specification).__name__)
        )

    # ----------------------------------------------------------------------------------------
    def lookup_class(self, class_type):
        """"""

        # This looks like a request to load a class at runtime?
        # The class type should be filename::classname.
        if "::" in class_type:

            RuntimeClass = import_class(class_type)

            return RuntimeClass

        raise NotFound("unable to get class for %s thing" % (class_type))



import logging
import re
from string import Template

import yaml

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.explain import explain2
from dls_utilpack.require import require
from dls_utilpack.substitute import substitute_string

# Exceptions.
from soakdb3_api.exceptions import NotFound

# Base class which maps flask requests to methods.
from soakdb3_lib.configurators.base import Base

logger = logging.getLogger(__name__)


thing_type = "soakdb3_lib.configurators.yaml"


class DottedTemplate(Template):
    """
    Allow tokens to have "." in them.
    """

    idpattern = r"(?a:[_.a-z][_.a-z0-9]*)"


class Yaml(Base):
    """
    Object representing a configurator which loads a yaml file.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None):
        self.__filename = "unknown filename"

        Base.__init__(self, thing_type, specification)

        type_specific_tbd = require(
            f"{callsign(self)} specification", specification, "type_specific_tbd"
        )

        self.__filename = require(
            f"{callsign(self)} specification type_specific_tbd",
            type_specific_tbd,
            "filename",
        )

        self.__removals = []
        self.__substitutions = {}
        self.__last_loaded_dict = {}

    # ----------------------------------------------------------------------------------------
    def callsign(self):
        return f"XchemBeConfigurator.Yaml {self.__filename}"

    # ----------------------------------------------------------------------------------------
    def require(self, keyword):
        parts = keyword.split(".")

        pointer = self.__last_loaded_dict

        for index, part in enumerate(parts):
            if not isinstance(pointer, dict):
                if index == len(parts) - 1:
                    raise NotFound(f"{callsign(self)} found non-dict at {keyword}")
                else:
                    raise NotFound(
                        f"{callsign(self)} found non-dict at %s when looking for %s"
                        % (".".join(parts[: index + 1]), keyword)
                    )

            if part in pointer:
                pointer = pointer[part]
            else:
                if index == len(parts) - 1:
                    raise NotFound(
                        f"{callsign(self)} is unable to find keyword {keyword}"
                    )
                else:
                    raise NotFound(
                        f"{callsign(self)} is unable to find keyword %s when looking for %s"
                        % (".".join(parts[: index + 1]), keyword)
                    )

        return pointer

    # ----------------------------------------------------------------------------------------
    def substitute(self, substitutions):
        self.__substitutions.update(substitutions)

    # ----------------------------------------------------------------------------------------
    def remove(self, removals):
        if isinstance(removals, list):
            self.__removals.extend(removals)
        else:
            self.__removals.append(removals)

    # ----------------------------------------------------------------------------------------
    async def load(self):
        try:
            with open(self.__filename, "r") as yaml_stream:
                loaded_string = yaml_stream.read()
        except Exception as exception:
            raise RuntimeError(f"unable to read {self.__filename}") from exception

        # Apply the substitutions to symbols in the loaded string.
        loaded_string = substitute_string(
            loaded_string, self.__substitutions, what=callsign(self)
        )

        try:
            loaded_dict = yaml.safe_load(loaded_string)
        except Exception as exception:
            raise RuntimeError(f"unable to parse {self.__filename}") from exception

        # Apply requested removals to the dict after parsing.
        self.__apply_removals(loaded_dict, self.__removals)

        self.__last_loaded_dict = loaded_dict

        return loaded_dict

    # ----------------------------------------------------------------------------------------
    def __apply_removals(self, loaded_dict, removals):
        for removal in removals:
            if removal in loaded_dict:
                loaded_dict.pop(removal)

    # ----------------------------------------------------------------------------------------
    def resolve(self, target):
        """
        Use the internal substitutions to resolve tokens in the given string.
        For example "${multi.level1.level2}".  See test_configurator for example.
        """

        if target is None:
            return None

        # Only a string can really have the dotted notation in it.
        # TODO: Extend configurator resolve to allow substitution in dicts.
        if not isinstance(target, str):
            return target

        try:
            # The regex which finds the ${tokens}.
            regex = re.compile(r"\$\{([^\}]+)")

            # Find all the tokens, eliciting the variable names.
            tokens = regex.findall(target)

            # Make the substitutions from within our current configuration state.
            substitutions = {}
            for index, token in enumerate(tokens):
                substitutions[token] = self.require(token)

            # Apply subtitutions to the target.
            return substitute_string(target, substitutions, what=callsign(self))
        except Exception as exception:
            raise RuntimeError(
                explain2(exception, f"{callsign(self)} substituting {target}")
            )
